from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class Result:
    @typechecked
    def __init__(self, attributes: list[tc.Variable | tuple[tc.Variable, str]]) -> None:
        """
        Parameters
        ----------
        attributes : list[Variable | tuple[Variable, str]]
        """
        assert (
            attributes is not None and len(attributes) > 0
        ), "Must return some variables / attributes."
        self.attributes = attributes

    @typechecked
    def __repr__(self) -> str:
        output = "["
        for x in self.attributes:
            if isinstance(x, tc.Variable):
                output += f"\\text{{{x.name}}}, "
            elif isinstance(x, tuple):
                variable, attribute = x
                output += f"\\text{{{variable.name}.{attribute}}}, "
        output = output[: -len(", ")]
        output += "]"
        return output



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.domain_calculus as dc


class Variable:
    @typechecked
    def __init__(self, name: str) -> None:
        self.name = name.lower()

    @typechecked
    def __repr__(self) -> str:
        return f"\\text{{{self.name}}}"



import pytest

from relational_calculus.tuple_calculus import *


def test_result_no_variables():
    with pytest.raises(Exception):
        result = Result([])


def test_double_typing_1():
    var = Variable("name", "Type")
    with pytest.raises(Exception):
        Exists(var, var).verify()


def test_double_typing_2():
    var = Variable("name", "Type")
    with pytest.raises(Exception):
        Exists(var, Forall(var, Equals(1, 1)), Equals(1, 1)).verify()


def test_renaming():
    var = Variable("name", "Type")
    formula = Or(Exists(var, Equals(1, 1)), Forall(var, Equals(1, 1)))
    assert len(formula.get_used_variables()) == 1
    formula = formula.to_normal_form()
    assert len(formula.get_used_variables()) == 2



import pytest

from relational_calculus.domain_calculus import *


def test_track_refs(session):
    r = session.execute("SELECT circuitRef FROM circuits;")
    solution = set(r.fetchall())
    assert len(solution) > 0

    cR = Variable("circuitRef")
    dc = DomainCalculus(
        Result([cR]), Tuple("circuits", [None, cR, None, None, None, None, None])
    )
    r = session.execute(dc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_tracks_germany(session):
    r = session.execute(
        "SELECT Name, Location FROM circuits WHERE Country = 'Germany';"
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    name = Variable("name")
    location = Variable("location")
    dc = DomainCalculus(
        Result([name, location]),
        Tuple("circuits", [None, None, name, location, "Germany", None, None]),
    )
    r = session.execute(dc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_tracks_northeast_southwest(session):
    r = session.execute(
        "SELECT * FROM circuits WHERE (Lat >= 0 AND Lng >= 0) OR (Lat < 0 AND Lng < 0);"
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    circuitId = Variable("id")
    circuitRef = Variable("red")
    name = Variable("ref")
    location = Variable("loc")
    country = Variable("country")
    lat = Variable("lat")
    lng = Variable("lng")
    dc = DomainCalculus(
        Result([circuitId, circuitRef, name, location, country, lat, lng]),
        And(
            Tuple(
                "circuits", [circuitId, circuitRef, name, location, country, lat, lng]
            ),
            Or(
                And(GreaterEquals(lat, 0), GreaterEquals(lng, 0)),
                And(LessThan(lat, 0), LessThan(lng, 0)),
            ),
        ),
    )
    r = session.execute(dc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_fastest_race_laps(session):
    r = session.execute(
        """SELECT circuits.name, drivers.forename, drivers.surname, lapTimes.milliseconds
        FROM circuits
        JOIN races ON circuits.circuitId = races.circuitId
        JOIN lapTimes ON races.raceId = lapTimes.raceId
        JOIN drivers ON lapTimes.driverId = drivers.driverId
        WHERE NOT EXISTS (
            SELECT *
            FROM lapTimes lT
            WHERE lT.raceId = lapTimes.raceId
            AND (lapTimes.driverID != lT.driverID OR lapTimes.lap != lT.lap)
            AND lapTimes.milliseconds < lT.milliseconds
        );"""
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    cN = Variable("cN")
    dF = Variable("dF")
    dS = Variable("dS")
    lM = Variable("lM")
    lL = Variable("lL")

    cI = Variable("cI")
    dI = Variable("dI")
    rI = Variable("rI")

    dI2 = Variable("dI2")
    lM2 = Variable("lM2")
    lL2 = Variable("lL2")
    dc = DomainCalculus(
        Result([cN, dF, dS, lM]),
        Exists(
            {cI, dI, rI},
            And(
                Tuple("races", [rI, None, None, cI, None, None, None]),
                And(
                    Tuple("circuits", [cI, None, cN, None, None, None, None]),
                    And(
                        Tuple("drivers", [dI, None, None, None, dF, dS, None, None]),
                        And(
                            Tuple("lapTimes", [rI, dI, lL, None, None, lM]),
                            Not(
                                Exists(
                                    {dI2, lM2, lL2},
                                    And(
                                        Tuple(
                                            "lapTimes", [rI, dI2, lL2, None, None, lM2]
                                        ),
                                        And(
                                            Or(
                                                Not(Equals(dI, dI2)),
                                                Not(Equals(lL, lL2)),
                                            ),
                                            LessThan(lM, lM2),
                                        ),
                                    ),
                                )
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )
    r = session.execute(dc.to_sql())
    result = set(r.fetchall())

    assert solution == result



import pytest

from relational_calculus.domain_calculus import *


def test_result_no_variables():
    with pytest.raises(Exception):
        result = Result(list())


def test_double_typing():
    var = Variable("x")
    with pytest.raises(Exception):
        Exists(var, Forall(var, Equals(1, 1)), Equals(1, 1)).verify()


def test_renaming():
    var = Variable("x")
    formula = Or(Exists(var, Equals(1, 1)), Forall(var, Equals(1, 1)))
    assert len(formula.get_used_variables()) == 1
    formula = formula.to_normal_form()
    assert len(formula.get_used_variables()) == 2



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class Exists(tc.Formula):
    """
    A class representing the exists quantifier in first order logic.
    """

    @typechecked
    def __init__(
        self, variable: tc.Variable | set[tc.Variable], child: tc.Formula
    ) -> None:
        """
        Parameters
        ----------
        variable : Variable | set[Variable]
            The variable(s) that should be quantified.
        child: Formula
            The sub-formula succeding this quantifier.
        """
        super().__init__([child])
        self.variable = variable
        if isinstance(self.variable, set) and len(self.variable) == 0:
            raise Exception("Must quantify at least one variable")

    @typechecked
    def __repr__(self) -> str:
        if isinstance(self.variable, set):
            forall_str = ""
            type_str = ""
            for var in self.variable:
                forall_str += f"{var.name}, "
                type_str += f"\\text{{{var.name}}} \\in \\text{{{var.type}}} \\land "
            forall_str = forall_str[: -len(", ")]
            type_str = type_str[: -len(" \\land ")]
        else:
            forall_str = self.variable.name
            type_str = (
                f"\\text{{{self.variable.name}}} \\in \\text{{{self.variable.type}}}"
            )

        return f"\\exists_{{{forall_str}}}({type_str} \\land ({self.children[0]}))"

    @typechecked
    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Exists)
            and self.variable == other.variable
            and self.children[0] == other.children[0]
        )

    @typechecked
    def expand_quantifiers(self) -> tc.Formula:
        if isinstance(self.variable, set):
            formula = self.children[0].expand_quantifiers()
            for variable in self.variable:
                formula = Exists(variable, formula)
            return formula
        else:
            return Exists(self.variable, self.children[0].expand_quantifiers())

    @typechecked
    def move_quantifiers_inwards(self) -> tc.Formula:
        assert isinstance(self.variable, tc.Variable)

        if not self.children[0].contains_variable(self.variable):
            return self.children[0].move_quantifiers_inwards()

        if isinstance(self.children[0], tc.And):
            if self.children[0].children[0].contains_variable(
                self.variable
            ) and not self.children[0].children[1].contains_variable(self.variable):
                return tc.And(
                    Exists(self.variable, self.children[0].children[0]),
                    self.children[0].children[1],
                ).move_quantifiers_inwards()
            elif self.children[0].children[1].contains_variable(
                self.variable
            ) and not self.children[0].children[0].contains_variable(self.variable):
                return tc.And(
                    self.children[0].children[0],
                    Exists(self.variable, self.children[0].children[1]),
                ).move_quantifiers_inwards()
            else:
                return Exists(
                    self.variable, self.children[0].move_quantifiers_inwards()
                )

        if isinstance(self.children[0], tc.Or):
            if self.children[0].children[0].contains_variable(
                self.variable
            ) and not self.children[0].children[1].contains_variable(self.variable):
                return tc.Or(
                    Exists(self.variable, self.children[0].children[0]),
                    self.children[0].children[1],
                ).move_quantifiers_inwards()
            elif self.children[0].children[1].contains_variable(
                self.variable
            ) and not self.children[0].children[0].contains_variable(self.variable):
                return tc.Or(
                    self.children[0].children[0],
                    Exists(self.variable, self.children[0].children[1]),
                ).move_quantifiers_inwards()
            else:
                return Exists(
                    self.variable, self.children[0].move_quantifiers_inwards()
                )

        if isinstance(self.children[0], tc.Not):
            # prefer exists over forall
            return Exists(self.variable, self.children[0].move_quantifiers_inwards())

        return Exists(self.variable, self.children[0].move_quantifiers_inwards())

    @typechecked
    def _move_same_type_quantifiers_outwards(
        self, move_exists: bool, move_forall: bool
    ) -> tc.Formula:
        return Exists(
            self.variable,
            self.children[0]._move_same_type_quantifiers_outwards(
                move_exists=True, move_forall=False
            ),
        )

    @typechecked
    def _to_sql(self) -> str:
        from_query = ""
        if isinstance(self.variable, set):
            for var in self.variable:
                from_query += f"{var.type} AS {var.name}, "
            from_query = from_query[: -len(", ")]
        else:
            from_query = f"{self.variable.type} AS {self.variable.name}"
        return f"EXISTS (SELECT NULL FROM {from_query} WHERE ({self.children[0]._to_sql()}))"

    @typechecked
    def flatten_quantifiers(self) -> tc.Formula:
        if isinstance(self.children[0], Exists):
            if isinstance(self.variable, set) and isinstance(
                self.children[0].variable, set
            ):
                new_variable = self.variable | self.children[0].variable
            elif isinstance(self.variable, set) and not isinstance(
                self.children[0].variable, set
            ):
                new_variable = self.variable | {self.children[0].variable}
            elif not isinstance(self.variable, set) and isinstance(
                self.children[0].variable, set
            ):
                new_variable = {self.variable} | self.children[0].variable
            else:
                new_variable = {self.variable, self.children[0].variable}
            return Exists(
                new_variable, self.children[0].children[0]
            ).flatten_quantifiers()
        else:
            return Exists(self.variable, self.children[0].flatten_quantifiers())

    @typechecked
    def _to_normal_form(self) -> tc.Formula:
        if isinstance(self.variable, set):
            formula = self.children[0]._to_normal_form()
            for var in self.variable:
                formula = Exists(var, formula)
            return formula
        return Exists(self.variable, self.children[0]._to_normal_form())

    @typechecked
    def contains_variable(self, variable: tc.Variable) -> bool:
        return (
            variable in self.variable
            if isinstance(self.variable, set)
            else self.variable == variable
        ) or self.children[0].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: tc.Variable) -> bool:
        return (
            variable in self.variable
            if isinstance(self.variable, set)
            else self.variable == variable
        ) or self.children[0].contains_variable_typing(variable)

    @typechecked
    def _rename_variable(
        self, old_variable: tc.Variable, new_variable: tc.Variable
    ) -> tc.Formula:
        if self.variable == old_variable:
            self.variable = new_variable
        if self.children[0] == old_variable:
            Exists(self.variable, new_variable)
        return Exists(
            self.variable, self.children[0]._rename_variable(old_variable, new_variable)
        )

    @typechecked
    def _prune_variable_atoms(self) -> tc.Formula | None:
        new_child = self.children[0]._prune_variable_atoms()
        if new_child is None:
            raise Exception("Invalid input: Cannot quantify an declared variable.")

        return Exists(self.variable, new_child)

    @typechecked
    def get_used_variables(self) -> set[tc.Variable]:
        child_vars = self.children[0].get_used_variables()
        if isinstance(self.variable, set):
            return child_vars.union(self.variable)
        return child_vars.union({self.variable})

    @typechecked
    def _check_variable_legality(self, variable: tc.Variable) -> bool:
        if self.variable != variable:
            return self.children[0]._check_variable_legality(variable)

        if self.children[0].contains_variable_typing(variable):
            raise Exception("Invalid input: Cannot quantify an declared variable.")

        return True



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.domain_calculus as dc


class Result:
    @typechecked
    def __init__(self, variables: list[dc.Variable]) -> None:
        """
        Parameters
        ----------
        variables : list[Variable]
            The variables that should be returned.
        """
        if len(variables) == 0:
            raise Exception("Must return at least one variable")
        self.variables = variables

    @typechecked
    def __repr__(self) -> str:
        output = ""
        for variable in self.variables:
            output += f"\\text{{{variable.name}}}, "
        output = output[:-2]  # remove last ', '
        return output



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class Not(tc.Formula):
    """
    A class representing the logical negation 'not'.
    """

    @typechecked
    def __init__(self, child: tc.Formula) -> None:
        """
        Parameters
        ----------
        child : Formula
            Subformula.
        """
        super().__init__([child])

    @typechecked
    def __repr__(self) -> str:
        if isinstance(self.children[0], tc.And | tc.Or):
            return f"\\neg({self.children[0]})"
        return f"\\neg {self.children[0]}"

    @typechecked
    def __eq__(self, other) -> bool:
        return isinstance(other, Not) and self.children[0] == other.children[0]

    @typechecked
    def expand_quantifiers(self) -> tc.Formula:
        return Not(self.children[0].expand_quantifiers())

    @typechecked
    def move_quantifiers_inwards(self) -> tc.Formula:
        return Not(self.children[0].move_quantifiers_inwards())

    @typechecked
    def _move_same_type_quantifiers_outwards(
        self, move_exists: bool, move_forall: bool
    ) -> tc.Formula:
        if move_exists:
            # prefer exists over forall
            if isinstance(self.children[0], tc.Forall):
                return tc.Exists(
                    self.children[0].variable, Not(self.children[0].children[0])
                )._move_same_type_quantifiers_outwards(
                    move_exists=True, move_forall=False
                )
            else:
                return Not(
                    self.children[0]._move_same_type_quantifiers_outwards(
                        move_exists=False, move_forall=move_forall
                    )
                )
        else:
            return Not(
                self.children[0]._move_same_type_quantifiers_outwards(
                    move_exists=False, move_forall=False
                )
            )

    @typechecked
    def _to_sql(self) -> str:
        if isinstance(self.children[0], tc.And | tc.Or):
            return f"NOT ({self.children[0]._to_sql()})"
        return f"NOT {self.children[0]._to_sql()}"

    @typechecked
    def flatten_quantifiers(self) -> tc.Formula:
        return Not(self.children[0].flatten_quantifiers())

    @typechecked
    def _to_normal_form(self) -> tc.Formula:
        if isinstance(self.children[0], Not):
            return self.children[0].children[0]._to_normal_form()

        if isinstance(self.children[0], tc.And):
            return tc.Or(
                Not(self.children[0].children[0]), Not(self.children[0].children[0])
            )._to_normal_form()

        if isinstance(self.children[0], tc.Or):
            return tc.And(
                Not(self.children[0].children[0]), Not(self.children[0].children[0])
            )._to_normal_form()

        if isinstance(self.children[0], tc.Exists):
            return tc.Forall(
                self.children[0].variable, Not(self.children[0].children[0])
            )._to_normal_form()

        if isinstance(self.children[0], tc.Forall):
            return tc.Exists(
                self.children[0].variable, Not(self.children[0].children[0])
            )._to_normal_form()

        return Not(self.children[0]._to_normal_form())

    @typechecked
    def contains_variable(self, variable: tc.Variable) -> bool:
        return self.children[0].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: tc.Variable) -> bool:
        return self.children[0].contains_variable_typing(variable)

    @typechecked
    def _rename_variable(
        self, old_variable: tc.Variable, new_variable: tc.Variable
    ) -> tc.Formula:
        if self.children[0] == old_variable:
            return Not(new_variable)
        return Not(self.children[0]._rename_variable(old_variable, new_variable))

    @typechecked
    def _prune_variable_atoms(self) -> tc.Formula | None:
        new_child = self.children[0]._prune_variable_atoms()
        if new_child is None:
            return None

        return Not(new_child)

    @typechecked
    def get_used_variables(self) -> set[tc.Variable]:
        return self.children[0].get_used_variables()

    @typechecked
    def _check_variable_legality(self, variable: tc.Variable) -> bool:
        if self.children[0].contains_variable_typing(variable):
            # reminder: we are in Normal Form
            raise Exception("The declaration of a variable cannot be negated!")

        return self.children[0]._check_variable_legality(variable)



import datetime

import pytest

from relational_calculus.tuple_calculus import *


def test_track_refs(session):
    r = session.execute("SELECT circuitRef FROM circuits;")
    solution = set(r.fetchall())
    assert len(solution) > 0

    c = Variable("circuit", "Circuits")
    tc = TupleCalculus(Result([c["circuitRef"]]), c)
    r = session.execute(tc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_tracks_germany(session):
    r = session.execute(
        "SELECT Name, Location FROM circuits WHERE Country = 'Germany';"
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    c = Variable("circuit", "Circuits")
    tc = TupleCalculus(
        Result([c["Name"], c["Location"]]),
        And(c, Equals((c, "Country"), "Germany")),
    )
    r = session.execute(tc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_tracks_northeast_southwest(session):
    r = session.execute(
        "SELECT * FROM circuits WHERE (Lat >= 0 AND Lng >= 0) OR (Lat < 0 AND Lng < 0);"
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    c = Variable("circuit", "Circuits")
    tc = TupleCalculus(
        Result([c]),
        And(
            c,
            Or(
                And(GreaterEquals((c, "Lat"), 0), GreaterEquals((c, "Lng"), 0)),
                And(LessThan((c, "Lat"), 0), LessThan((c, "Lng"), 0)),
            ),
        ),
    )
    r = session.execute(tc.to_sql())
    result = set(r.fetchall())

    assert solution == result


def test_fastest_race_laps(session):
    r = session.execute(
        """SELECT circuits.name, drivers.forename, drivers.surname, lapTimes.milliseconds
        FROM circuits
        JOIN races ON circuits.circuitId = races.circuitId
        JOIN lapTimes ON races.raceId = lapTimes.raceId
        JOIN drivers ON lapTimes.driverId = drivers.driverId
        WHERE NOT EXISTS (
            SELECT *
            FROM lapTimes lT
            WHERE lT.raceId = lapTimes.raceId
            AND (lapTimes.driverID != lT.driverID OR lapTimes.lap != lT.lap)
            AND lapTimes.milliseconds < lT.milliseconds
        );"""
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    circuit = Variable("circuit", "Circuits")
    race = Variable("race", "Races")
    lapTime = Variable("lapTime", "lapTimes")
    driver = Variable("driver", "Drivers")

    c2 = Variable("c2", "Circuits")
    r2 = Variable("r2", "Races")
    lT2 = Variable("lT2", "lapTimes")
    d2 = Variable("d2", "Drivers")

    tc = TupleCalculus(
        Result(
            [
                circuit["name"],
                driver["forename"],
                driver["surname"],
                lapTime["milliseconds"],
            ]
        ),
        And(
            And(And(circuit, lapTime), driver),
            Exists(
                race,
                And(
                    # Join On
                    And(
                        Equals((circuit, "circuitId"), (race, "circuitId")),
                        And(
                            Equals((race, "raceId"), (lapTime, "raceId")),
                            Equals((lapTime, "driverId"), (driver, "driverId")),
                        ),
                    ),
                    # Fastest Time
                    Forall(
                        lT2,
                        Or(
                            Not(
                                And(
                                    Equals((lT2, "raceId"), (lapTime, "raceId")),
                                    Not(
                                        And(
                                            Equals(
                                                (lT2, "driverId"), (lapTime, "driverId")
                                            ),
                                            Equals((lT2, "lap"), (lapTime, "lap")),
                                        )
                                    ),
                                )
                            ),
                            LessEquals(
                                (lT2, "milliseconds"), (lapTime, "milliseconds")
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )
    r = session.execute(tc.to_sql())
    result = set(r.fetchall())

    assert solution == result



from __future__ import annotations

from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class Exists(dc.Formula):
    @typechecked
    def __init__(
        self, variable: dc.Variable | set[dc.Variable], child: dc.Formula
    ) -> None:
        super().__init__([child])
        self.variable = variable
        if isinstance(self.variable, set) and len(self.variable) == 0:
            raise Exception("Must quantify at least one variable")

    @typechecked
    def __repr__(self) -> str:
        if isinstance(self.variable, set):
            forall_str = ""
            for var in self.variable:
                forall_str += f"{var.name}, "
            forall_str = forall_str[: -len(", ")]
        else:
            forall_str = self.variable.name

        return f"\\exists_{{{forall_str}}}({self.children[0]})"

    @typechecked
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Exists)
            and self.variable == other.variable
            and self.children[0] == other.children[0]
        )

    @typechecked
    def _to_sql(self) -> Optional[str]:
        return None

    @typechecked
    def _to_normal_form(self) -> dc.Formula:
        if isinstance(self.variable, set):
            formula = self.children[0]._to_normal_form()
            for var in self.variable:
                formula = Exists(var, formula)
            return formula
        return Exists(self.variable, self.children[0]._to_normal_form())

    @typechecked
    def contains_variable(self, variable: dc.Variable) -> bool:
        return (
            variable in self.variable
            if isinstance(self.variable, set)
            else self.variable == variable
        ) or self.children[0].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        return False

    @typechecked
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        return self.contains_variable(variable) or self.children[
            0
        ].contains_variable_quantification(variable)

    @typechecked
    def _rename_variable(
        self, old_variable: dc.Variable, new_variable: dc.Variable
    ) -> dc.Formula:
        if isinstance(self.variable, set):
            new_set = set(self.variable)
            for var in self.variable:
                if var == old_variable:
                    new_set.remove(var)
                    new_set.add(new_variable)
            return Exists(
                new_set, self.children[0]._rename_variable(old_variable, new_variable)
            )
        else:
            if self.variable == old_variable:
                return Exists(
                    new_variable,
                    self.children[0]._rename_variable(old_variable, new_variable),
                )
            else:
                return Exists(
                    self.variable,
                    self.children[0]._rename_variable(old_variable, new_variable),
                )

    @typechecked
    def _prune_tuple_atoms(self) -> dc.Formula | None:
        new_child = self.children[0]._prune_tuple_atoms()
        if new_child is None:
            return None
        if isinstance(self.variable, set):
            new_set = set(self.variable)
            for var in self.variable:
                if not new_child.contains_variable_typing(var):
                    new_set.remove(var)
            if len(new_set) == 0:
                return new_child
            return Exists(new_set, new_child)
        else:
            if not new_child.contains_variable_typing(self.variable):
                return new_child
            return Exists(self.variable, new_child)

    @typechecked
    def get_used_variables(self) -> set[dc.Variable]:
        child_vars = self.children[0].get_used_variables()
        if isinstance(self.variable, set):
            return child_vars.union(self.variable)
        return child_vars.union({self.variable})

    @typechecked
    def get_used_tuples(self) -> set[dc.Tuple]:
        return self.children[0].get_used_tuples()

    @typechecked
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        if self.variable != variable:
            return self.children[0]._check_variable_legality(variable)

        if self.children[0].contains_variable_quantification(variable):
            raise Exception("Invalid input: Cannot quantify an declared variable.")

        return True



from __future__ import annotations

from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class LessThan(dc.Formula):
    @typechecked
    def __init__(
        self,
        left: dc.PRIMITIVE_TYPES | dc.Variable,
        right: dc.PRIMITIVE_TYPES | dc.Variable,
    ) -> None:
        super().__init__([])
        self.left = left
        self.right = right

    @typechecked
    def __repr__(self) -> str:
        return f"{self.left} < {self.right}"

    @typechecked
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, LessThan)
            and self.left == other.left
            and self.right == other.right
        )

    @typechecked
    def _to_sql(self) -> Optional[str]:
        if isinstance(self.left, dc.Variable):
            left_string = f'"{self.left.name}"'
        else:
            left_string = f"{str(self.left)}"
        if isinstance(self.right, dc.Variable):
            right_string = f'"{self.right.name}"'
        else:
            right_string = f"{str(self.right)}"

        return f"{left_string} < {right_string}"

    @typechecked
    def _to_normal_form(self) -> dc.Formula:
        return self

    @typechecked
    def contains_variable(self, variable: dc.Variable) -> bool:
        return self.left == variable or self.right == variable

    @typechecked
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        return False

    @typechecked
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        return False

    @typechecked
    def _rename_variable(
        self, old_variable: dc.Variable, new_variable: dc.Variable
    ) -> dc.Formula:
        if self.left == old_variable:
            self.left = new_variable
        if self.right == old_variable:
            self.right = new_variable
        return self

    @typechecked
    def _prune_tuple_atoms(self) -> dc.Formula | None:
        return self

    @typechecked
    def get_used_variables(self) -> set[dc.Variable]:
        variables = set()

        if isinstance(self.left, dc.Variable):
            variables.add(self.left)

        if isinstance(self.right, dc.Variable):
            variables.add(self.right)

        return variables

    @typechecked
    def get_used_tuples(self) -> set[dc.Tuple]:
        return set()

    @typechecked
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        return True



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class GreaterThan(tc.Formula):
    """
    A class representing the comparison '>'.
    """

    @typechecked
    def __init__(
        self,
        left: tuple[tc.Variable, str] | tc.PRIMITIVE_TYPES,
        right: tuple[tc.Variable, str] | tc.PRIMITIVE_TYPES,
    ) -> None:
        """
        Parameters
        ----------
        left : tuple[Variable, str] | PRIMITIVE_TYPES (such as str, int, float)
            Left-hand side of the comparison.
        right : tuple[Variable, str] | PRIMITIVE_TYPES (such as str, int, float)
            Right-hand side of the comparison.
        """
        super().__init__([])
        self.left = left
        self.right = right

    @typechecked
    def __repr__(self) -> str:
        if isinstance(self.left, tuple):  # (variable, attr_name)
            left_string = f"\\text{{{self.left[0].name}.{self.left[1]}}}"
        else:
            left_string = f"\\text{{{self.left}}}"
        if isinstance(self.right, tuple):  # (variable, attr_name)
            right_string = f"\\text{{{self.right[0].name}.{self.right[1]}}}"
        else:
            right_string = f"\\text{{{self.right}}}"
        return f"{left_string} > {right_string}"

    @typechecked
    def __eq__(self, other) -> bool:
        return (
            isinstance(other, GreaterThan)
            and self.left == other.left
            and self.right == other.right
        )

    @typechecked
    def expand_quantifiers(self) -> tc.Formula:
        return self

    @typechecked
    def move_quantifiers_inwards(self) -> tc.Formula:
        return self

    @typechecked
    def _move_same_type_quantifiers_outwards(
        self, move_exists: bool, move_forall: bool
    ) -> tc.Formula:
        return self

    @typechecked
    def _to_sql(self) -> str:
        if isinstance(self.left, tuple):  # (variable, attr_name)
            left_string = f"{self.left[0].name}.{self.left[1]}"
        else:
            left_string = f"{self.left}"
        if isinstance(self.right, tuple):  # (variable, attr_name)
            right_string = f"{self.right[0].name}.{self.right[1]}"
        else:
            right_string = f"{self.right}"
        return f"{left_string} > {right_string}"

    @typechecked
    def flatten_quantifiers(self) -> tc.Formula:
        return self

    @typechecked
    def _to_normal_form(self) -> tc.Formula:
        return self

    @typechecked
    def contains_variable(self, variable: tc.Variable) -> bool:
        return (isinstance(self.left, tuple) and self.left[0] == variable) or (
            isinstance(self.right, tuple) and self.right[0] == variable
        )

    @typechecked
    def contains_variable_typing(self, variable: tc.Variable) -> bool:
        return False

    @typechecked
    def _rename_variable(
        self, old_variable: tc.Variable, new_variable: tc.Variable
    ) -> tc.Formula:
        if isinstance(self.left, tuple) and self.left[0] == old_variable:
            self.left = (new_variable, self.left[1])
        if isinstance(self.right, tuple) and self.right[0] == old_variable:
            self.right = (new_variable, self.right[1])
        return self

    @typechecked
    def _prune_variable_atoms(self) -> tc.Formula | None:
        return self

    @typechecked
    def get_used_variables(self) -> set[tc.Variable]:
        variables = set()

        if isinstance(self.left, tuple):
            variables.add(self.left[0])

        if isinstance(self.right, tuple):
            variables.add(self.right[0])

        return variables

    @typechecked
    def _check_variable_legality(self, variable: tc.Variable) -> bool:
        return True



from __future__ import annotations

from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class And(dc.Formula):
    @typechecked
    def __init__(self, child1: dc.Formula, child2: dc.Formula) -> None:
        super().__init__([child1, child2])

    @typechecked
    def __repr__(self) -> str:
        left = self.children[0]
        right = self.children[1]

        if isinstance(self.children[0], And):
            left = (
                f"{self.children[0].children[0]} \\land {self.children[0].children[1]}"
            )
        elif not isinstance(self.children[0], dc.Not | dc.ATOM_TYPES | dc.Tuple):
            left = f"({left})"

        if isinstance(self.children[1], And):
            right = (
                f"{self.children[1].children[0]} \\land {self.children[1].children[1]}"
            )
        elif not isinstance(self.children[1], dc.Not | dc.ATOM_TYPES | dc.Tuple):
            right = f"({right})"

        return f"{left} \\land {right}"

    @typechecked
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, And)
            and self.children[0] == other.children[0]
            and self.children[1] == other.children[1]
        )

    @typechecked
    def _to_sql(self) -> Optional[str]:
        return f"({self.children[0].to_sql()} AND {self.children[1].to_sql()})"

    @typechecked
    def _to_normal_form(self) -> dc.Formula:
        if isinstance(self.children[0], dc.Exists):
            self.children[0] = self.children[0]._to_normal_form()
            variable = self.children[0].variable
            if self.children[1].contains_variable(variable):
                return And(
                    self.children[0].rename_variable(variable), self.children[1]
                )._to_normal_form()
            else:
                return dc.Exists(
                    variable,
                    And(
                        self.children[0].children[0], self.children[1]
                    )._to_normal_form(),
                )

        if isinstance(self.children[0], dc.Forall):
            self.children[0] = self.children[0]._to_normal_form()
            variable = self.children[0].variable
            if self.children[1].contains_variable(variable):
                return And(
                    self.children[0].rename_variable(variable), self.children[1]
                )._to_normal_form()
            else:
                return dc.Forall(
                    variable,
                    And(
                        self.children[0].children[0], self.children[1]
                    )._to_normal_form(),
                )

        if isinstance(self.children[1], dc.Exists):
            self.children[1] = self.children[1]._to_normal_form()
            variable = self.children[1].variable
            if self.children[0].contains_variable(variable):
                return And(
                    self.children[0], self.children[1].rename_variable(variable)
                )._to_normal_form()
            else:
                return dc.Exists(
                    variable,
                    And(
                        self.children[0], self.children[1].children[0]
                    )._to_normal_form(),
                )

        if isinstance(self.children[1], dc.Forall):
            self.children[1] = self.children[1]._to_normal_form()
            variable = self.children[1].variable
            if self.children[0].contains_variable(variable):
                return And(
                    self.children[0], self.children[1].rename_variable(variable)
                )._to_normal_form()
            else:
                return dc.Forall(
                    variable,
                    And(
                        self.children[0], self.children[1].children[0]
                    )._to_normal_form(),
                )

        return And(
            self.children[0]._to_normal_form(), self.children[1]._to_normal_form()
        )

    @typechecked
    def contains_variable(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable(variable) or self.children[
            1
        ].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable_typing(variable) or self.children[
            1
        ].contains_variable_typing(variable)

    @typechecked
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable_quantification(
            variable
        ) or self.children[1].contains_variable_quantification(variable)

    @typechecked
    def _rename_variable(
        self, old_variable: dc.Variable, new_variable: dc.Variable
    ) -> dc.Formula:
        return And(
            self.children[0]._rename_variable(old_variable, new_variable),
            self.children[1]._rename_variable(old_variable, new_variable),
        )

    @typechecked
    def _prune_tuple_atoms(self) -> dc.Formula | None:
        left = self.children[0]._prune_tuple_atoms()
        right = self.children[1]._prune_tuple_atoms()

        if left is None and right is None:
            return None
        elif left is None:
            return right
        elif right is None:
            return left
        else:
            return And(left, right)

    @typechecked
    def get_used_variables(self) -> set[dc.Variable]:
        return (
            self.children[0]
            .get_used_variables()
            .union(self.children[1].get_used_variables())
        )

    @typechecked
    def get_used_tuples(self) -> set[dc.Tuple]:
        return (
            self.children[0].get_used_tuples().union(self.children[1].get_used_tuples())
        )

    @typechecked
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        if self.children[0].contains_variable_quantification(
            variable
        ) and self.children[1].contains_variable_quantification(variable):
            # cannot happen due to renaming of variables
            raise Exception("Internal Error")

        if self.children[0].contains_variable_quantification(
            variable
        ) and self.children[1].contains_variable(variable):
            raise Exception("Each variable must be declared before it is used")

        if self.children[0].contains_variable(variable) and self.children[
            1
        ].contains_variable_quantification(variable):
            raise Exception("Each variable must be declared before it is used")

        return self.children[0]._check_variable_legality(variable) and self.children[
            1
        ]._check_variable_legality(variable)



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.domain_calculus as dc
from relational_calculus.domain_calculus.formulas.quantifiers.Forall import Forall


class DomainCalculus:
    @typechecked
    def __init__(self, result: dc.Result, formula: dc.Formula) -> None:
        """
        Parameters
        ----------
        result : Result
            The result that should be returned.
        formula : Formula
            The formula that every result has to satisfy.
        """
        self.result = result
        self.formula = formula

    @typechecked
    def verify(self) -> bool:
        """
        Verify the correctness of the result/formula combination (e.g. one must specify the type of a variable before returning it).

        Returns
        -------
        bool
            True if the result/formula combination is correct, False otherwise.
        """
        assert self.formula is not None
        assert self.result is not None

        check = self.formula.verify()

        # Get Types of return variables
        variables = self.result.variables

        for variable in variables:
            check = check and self.formula.check_variable_legality(variable)

        return check

    @typechecked
    def __repr__(self) -> str:
        return f"\\{{[{self.result}] \\vert {self.formula}\\}}"

    @typechecked
    def to_sql(self) -> str:
        """
        Convert the domain calculus to a SQL query.

        Returns
        -------
        str
            The SQL query.
        """
        # Select Statement
        select_query = "SELECT DISTINCT "
        for variable in self.result.variables:
            select_query += f'"{variable.name}", '
        select_query = select_query[: -len(", ")]  # remove last ', '

        formula = self.formula.to_normal_form()
        used_tuples = formula.get_used_tuples()

        # Map Tuples to new Table Names
        id = 1
        tuple_mapping = dict()
        for tuple in used_tuples:
            tuple_mapping[tuple] = f"table_{id}"
            id += 1

        # Determine Tuples containing return variables
        result_tuples = set()
        for variable in self.result.variables:
            for tuple in used_tuples:
                if variable in tuple.variables:
                    result_tuples.add(tuple)

        # Determine other tuples order and type
        tuple_quantification = dict()
        tuple_order = list()
        while isinstance(formula, dc.QUANTIFIER_TYPES):
            variables = set()
            if isinstance(formula.variable, set):
                variables = formula.variable
            else:
                variables.add(formula.variable)
            for variable in variables:
                for tuple in used_tuples:
                    if tuple in result_tuples:
                        continue
                    if variable in tuple.variables:
                        if tuple not in tuple_order:
                            tuple_order.append(tuple)
                            tuple_quantification[tuple] = isinstance(formula, Forall)
                        else:
                            assert tuple in tuple_quantification.keys()
                            tuple_quantification[tuple] = (
                                isinstance(formula, Forall)
                                or tuple_quantification[tuple]
                            )
            formula = formula.children[0]

        # Create regions of adjecent tuples of the same type (Exists, Forall)
        tuple_regions = list()
        current_region = list()
        for tuple in tuple_order:
            if len(current_region) == 0:
                current_region.append(tuple)
            elif tuple_quantification[tuple] == tuple_quantification[current_region[0]]:
                current_region.append(tuple)
            else:
                tuple_regions.append(current_region)
                current_region = list()
                current_region.append(tuple)
        if len(current_region) > 0:
            tuple_regions.append(current_region)

        # With-As Statements
        with_query = "WITH "
        for tuple in used_tuples:
            __quotation_mark = '"'  # cannot use backslash in f''-string
            with_query += f'{tuple_mapping[tuple]}({",".join(list(map(lambda var: __quotation_mark + var.name + __quotation_mark, tuple.variables)))}) AS (SELECT * FROM {tuple.type}), '
        with_query = with_query[: -len(", ")]

        # From Statement
        from_query = " FROM "
        for tuple in result_tuples:
            from_query += f"{tuple_mapping[tuple]} NATURAL JOIN "
        from_query = from_query[: -len(" NATURAL JOIN ")]

        # Where Statement
        where_sub_query = formula.to_sql()
        where_query = ""
        if where_sub_query is not None and where_sub_query != "":
            where_query = " WHERE "
            bracket_counter = 0
            previous_tuples = list(result_tuples)
            if len(tuple_regions) > 0:
                for region in tuple_regions:
                    _from_sub_query = "FROM "
                    for tuple in region:
                        _from_sub_query += f"{tuple_mapping[tuple]} NATURAL JOIN "
                    _from_sub_query = _from_sub_query[: -len(" NATURAL JOIN ")]

                    _where_sub_query = "WHERE"
                    for previous_tuple in previous_tuples:
                        for tuple in region:
                            for variable in previous_tuple.variables:
                                if variable in tuple.variables:
                                    _where_sub_query += f" {tuple_mapping[previous_tuple]}.{variable.name} = {tuple_mapping[tuple]}.{variable.name} AND"

                    if tuple_quantification[region[0]]:
                        where_query += f"NOT EXISTS (SELECT NULL {_from_sub_query} {_where_sub_query} NOT ("
                        bracket_counter += 2
                    else:
                        where_query += (
                            f"EXISTS (SELECT NULL {_from_sub_query} {_where_sub_query} "
                        )
                        bracket_counter += 1
                    previous_tuples.extend(region)

            where_query += where_sub_query
            where_query += ")" * bracket_counter

        return f"{with_query} {select_query} {from_query} {where_query}"



import pytest

from relational_calculus.tuple_calculus import *


def test_alternative_syntax():
    var = Variable("name", "Type")
    assert var["a"] == (var, "a")


def test_alternative_syntax_error_multiple():
    var = Variable("name", "Type")
    with pytest.raises(Exception):
        var["a", "b"]


def test_alternative_syntax_error_type():
    var = Variable("name", "Type")
    with pytest.raises(Exception):
        var[1]


def test_fastest_race_laps(session):
    r = session.execute(
        """SELECT circuits.name, drivers.forename, drivers.surname, lapTimes.milliseconds
        FROM circuits
        JOIN races ON circuits.circuitId = races.circuitId
        JOIN lapTimes ON races.raceId = lapTimes.raceId
        JOIN drivers ON lapTimes.driverId = drivers.driverId
        WHERE NOT EXISTS (
            SELECT *
            FROM lapTimes lT
            WHERE lT.raceId = lapTimes.raceId
            AND (lapTimes.driverID != lT.driverID OR lapTimes.lap != lT.lap)
            AND lapTimes.milliseconds < lT.milliseconds
        );"""
    )
    solution = set(r.fetchall())
    assert len(solution) > 0

    circuit = Variable("circuit", "Circuits")
    race = Variable("race", "Races")
    lapTime = Variable("lapTime", "lapTimes")
    driver = Variable("driver", "Drivers")

    c2 = Variable("c2", "Circuits")
    r2 = Variable("r2", "Races")
    lT2 = Variable("lT2", "lapTimes")
    d2 = Variable("d2", "Drivers")

    tc = TupleCalculus(
        Result(
            [
                circuit["name"],
                driver["forename"],
                driver["surname"],
                lapTime["milliseconds"],
            ]
        ),
        And(
            And(And(circuit, lapTime), driver),
            Exists(
                race,
                And(
                    # Join On
                    And(
                        Equals(circuit["circuitId"], race["circuitId"]),
                        And(
                            Equals(race["raceId"], lapTime["raceId"]),
                            Equals(lapTime["driverId"], driver["driverId"]),
                        ),
                    ),
                    # Fastest Time
                    Forall(
                        lT2,
                        Or(
                            Not(
                                And(
                                    Equals(lT2["raceId"], lapTime["raceId"]),
                                    Not(
                                        And(
                                            Equals(lT2["driverId"], driver["driverId"]),
                                            Equals(lT2["lap"], lapTime["lap"]),
                                        )
                                    ),
                                )
                            ),
                            LessEquals(lT2["milliseconds"], lapTime["milliseconds"]),
                        ),
                    ),
                ),
            ),
        ),
    )
    r = session.execute(tc.to_sql())
    result = set(r.fetchall())

    assert solution == result



from __future__ import annotations

from abc import abstractclassmethod

from docstring_inheritance import NumpyDocstringInheritanceMeta
from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class Formula(metaclass=NumpyDocstringInheritanceMeta):
    """
    An abstract class used to represent a first order logic formula.
    This class has many different implementations, together forming a recursive data-structure to hold the formula.
    """

    @typechecked
    def __init__(self, children: list[Formula]) -> None:
        """
        Parameters
        ----------
        children : list[Formula]
            Each implementation of formula can hold a list of children that it is composed of.
        """
        self.children = children

    @typechecked
    @abstractclassmethod
    def __repr__(self) -> str:
        """
        Returs a string representation of a given formula formatted in Latex Math Mode.

        Returns
        -------
        str
            A string representation of the formula formatted in Latex Math Mode.
        """
        pass

    @typechecked
    @abstractclassmethod
    def __eq__(self, other: Any) -> bool:
        """
        Check for equality of syntax, not semantics.

        Returns
        -------
        bool
            If the objects are equal.
        """
        pass

    @typechecked
    def verify(self) -> bool:
        """
        Check if the formula is valid.

        Returns
        -------
        bool
            If the formula is valid.
        """
        formula = self.to_normal_form()
        variables = formula.get_used_variables()
        check = True
        for variable in variables:
            check = check and formula.check_variable_legality(variable)
        return check

    @typechecked
    def to_sql(self) -> Optional[str]:
        """
        Used by DomainCalculus::to_sql to convert the domain calculus specification into an equivalent SQL formula.

        Returns
        -------
        str
            The sub_query in SQL generated by this formula.
        """
        formula = self.prune_tuple_atoms()
        if formula is not None:
            return formula._to_sql()
        return ""

    @typechecked
    @abstractclassmethod
    def _to_sql(self) -> Optional[str]:
        """
        Used by Formula::to_sql to convert the domain calculus specification into an equivalent SQL formula.

        Returns
        -------
        str
            The sub_query in SQL generated by this formula.
        """
        pass

    @typechecked
    def to_normal_form(self) -> Formula:
        """
        Converts a given formula into an semantically equivalent Formula close to Prenex-Normalform.

        Returns
        -------
        Formula
            The semantically equivalent Formula close to Prenex-Normalform.
        """
        old_formula = self
        new_formula = old_formula._to_normal_form()
        while new_formula != old_formula:
            old_formula = new_formula
            new_formula = old_formula._to_normal_form()
        return new_formula

    @typechecked
    @abstractclassmethod
    def _to_normal_form(self) -> Formula:
        """
        Converts a given formula into an semantically equivalent Formula close to Prenex-Normalform.

        Returns
        -------
        Formula
            The semantically equivalent Formula close to Prenex-Normalform.
        """
        pass

    @typechecked
    @abstractclassmethod
    def contains_variable(self, variable: dc.Variable) -> bool:
        """
        Checks whether a given formula contains a given variable.

        Parameters
        ----------
        variable : Variable
            The variable to check in the formula.
        """
        pass

    @typechecked
    @abstractclassmethod
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        """
        Checks whether a given variable is being used in a tuple in a given formula.

        Parameters
        ----------
        variable : Variable
            The variable to check in the formula.
        """
        pass

    @typechecked
    @abstractclassmethod
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        """
        Checks whether a given variable is being quantified in a given formula..

        Parameters
        ----------
        variable : Variable
            The variable to check in the formula.
        """
        pass

    _rename_index = 0

    @typechecked
    def rename_variable(self, variable: dc.Variable) -> Formula:
        """
        Renames a variable to an unused name and replaces all occurences of the old variable with the new one.

        Parameters
        ----------
        variable : Variable
            The variable to rename.
        """
        Formula._rename_index += 1
        new_variable = dc.Variable(
            variable.name + str(Formula._rename_index),
        )
        return self._rename_variable(variable, new_variable)

    @typechecked
    @abstractclassmethod
    def _rename_variable(
        self, old_variable: dc.Variable, new_variable: dc.Variable
    ) -> Formula:
        """
        Replaces all old occurences of an old variable with the new variable.

        Parameters
        ----------
        old_variable : Variable
            The variable to replace.
        new_variable : Variable
            The variable to replace with.
        """
        pass

    @typechecked
    def prune_tuple_atoms(self) -> Formula | None:
        """
        Converts a given formula into a formula stripped of Tuple-atoms.

        Returns
        -------
        Formula
            The formula stripped of Tuple-atoms.
        """
        formula = self
        new_formula = formula._prune_tuple_atoms()
        while formula != new_formula and new_formula is not None:
            formula = new_formula
            new_formula = formula._prune_tuple_atoms()
        return new_formula

    @typechecked
    @abstractclassmethod
    def _prune_tuple_atoms(self) -> Formula | None:
        """
        Converts a given formula into a formula stripped of Tuple-atoms.

        Returns
        -------
        Formula
            The formula stripped of Tuple-atoms.
        """
        pass

    @typechecked
    @abstractclassmethod
    def get_used_variables(self) -> set[dc.Variable]:
        """
        Returns a set of variables used in a given formula.

        Returns
        -------
        set[Variable]
            A set of variables used in the formula.
        """
        pass

    @typechecked
    @abstractclassmethod
    def get_used_tuples(self) -> set[dc.Tuple]:
        """
        Returns a set of tuples used in a given formula.

        Returns
        -------
        set[Variable]
            A set of variables used in the formula.
        """
        pass

    @typechecked
    def check_variable_legality(self, variable: dc.Variable) -> bool:
        """
        Checks if a given variable is used correctly in the given formula.

        Parameters
        ----------
        variable : Variable
            The variable to check the legal use of.

        Returns
        -------
        true
            If the variable's usage was legal.
        false
            Otherwise.

        Raises
        ------
        Exception
            If the variable was not used legally. The exception message also holds further explanation.
        """
        return self.contains_variable_typing(
            variable
        ) and self.to_normal_form()._check_variable_legality(variable)

    @typechecked
    @abstractclassmethod
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        """
        Checks if a given variable is used correctly in the given formula.

        Parameters
        ----------
        variable : Variable
            The variable to check the legal use of.

        Returns
        -------
        true
            If the variable's usage was legal.
        false
            Otherwise.

        Raises
        ------
        Exception
            If the variable was not used legally. The exception message also holds further explanation.
        """
        pass



from relational_calculus.domain_calculus.Variable import Variable

from relational_calculus.domain_calculus.formulas.Formula import Formula
from relational_calculus.domain_calculus.formulas.Tuple import Tuple

from relational_calculus.domain_calculus.formulas.atoms.Equals import Equals
from relational_calculus.domain_calculus.formulas.atoms.GreaterEquals import (
    GreaterEquals,
)
from relational_calculus.domain_calculus.formulas.atoms.GreaterThan import GreaterThan
from relational_calculus.domain_calculus.formulas.atoms.LessEquals import LessEquals
from relational_calculus.domain_calculus.formulas.atoms.LessThan import LessThan

from relational_calculus.domain_calculus.formulas.connectives.And import And
from relational_calculus.domain_calculus.formulas.connectives.Or import Or
from relational_calculus.domain_calculus.formulas.connectives.Not import Not

from relational_calculus.domain_calculus.formulas.quantifiers.Exists import Exists
from relational_calculus.domain_calculus.formulas.quantifiers.Forall import Forall

from relational_calculus.domain_calculus.Result import Result
from relational_calculus.domain_calculus.DomainCalculus import DomainCalculus

PRIMITIVE_TYPES = int | float | str
ATOM_TYPES = Equals | GreaterEquals | GreaterThan | LessEquals | LessThan
CONNECTIVE_TYPES = And | Or | Not
QUANTIFIER_TYPES = Exists | Forall



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class TupleCalculus:
    @typechecked
    def __init__(self, result: tc.Result, formula: tc.Formula) -> None:
        """
        Parameters
        ----------
        result : Result
            The result that should be returned.
        formula : Formula
            The formula that every result has to satisfy.
        """
        self.result = result
        self.formula = formula

    @typechecked
    def verify(self) -> bool:
        """
        Verify the correctness of the result/formula combination (e.g. one must specify the type of a variable before returning it).

        Returns
        -------
        bool
            True if the result/formula combination is correct, False otherwise.
        """
        assert self.formula is not None
        assert self.result is not None

        check = self.formula.verify()

        # Get Types of return variables
        variables = set()
        for x in self.result.attributes:
            if isinstance(x, tc.Variable):
                variables.add(x)
            elif isinstance(x, tuple):
                variable, _ = x
                variables.add(variable)

        for variable in variables:
            check = check and self.formula.check_variable_legality(variable)

        return check

    @typechecked
    def __repr__(self) -> str:
        return f"\\{{{self.result} \\vert {self.formula}\\}}"

    @typechecked
    def to_sql(self, optimize: bool = True) -> str:
        """
        Convert the domain calculus to a SQL query.

        Parameters
        ----------
        optimize : bool, optional
            If the query should be optimized, by default True

        Returns
        -------
        str
            The SQL query.
        """
        # Get Types of return variables
        variables = set()
        for x in self.result.attributes:
            if isinstance(x, tc.Variable):
                variables.add(x)
            elif isinstance(x, tuple):
                variable, _ = x
                variables.add(variable)

        select_query = ""
        for x in self.result.attributes:
            if isinstance(x, tc.Variable):
                select_query += f"{x.name}.*, "
            elif isinstance(x, tuple):
                variable, attribute = x
                select_query += f"{variable.name}.{attribute}, "

        select_query = select_query[: -len(", ")]

        from_variables = variables
        formula = self.formula
        if optimize:
            formula = formula.optimize()
            if formula is not None:
                if isinstance(formula, tc.Exists):
                    if isinstance(formula.variable, set):
                        from_variables = from_variables.union(formula.variable)
                    else:
                        from_variables.add(formula.variable)
                    formula = formula.children[0]

        from_query = ""
        for variable in from_variables:
            from_query += f"{variable.type} {variable.name}, "
        from_query = from_query[: -len(", ")]

        query = f"SELECT DISTINCT {select_query} FROM {from_query}"

        where_query = None
        if formula is not None:
            where_query = formula.to_sql()

        if where_query is not None and where_query != "":
            query += f" WHERE {where_query}"
        return query



from __future__ import annotations

from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class Not(dc.Formula):
    @typechecked
    def __init__(self, child: dc.Formula) -> None:
        super().__init__([child])

    @typechecked
    def __repr__(self) -> str:
        if isinstance(self.children[0], dc.And | dc.Or):
            return f"\\neg({self.children[0]})"
        return f"\\neg {self.children[0]}"

    @typechecked
    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Not) and self.children[0] == other.children[0]

    @typechecked
    def _to_sql(self) -> Optional[str]:
        return f"(NOT {self.children[0].to_sql()})"

    @typechecked
    def _to_normal_form(self) -> dc.Formula:
        if isinstance(self.children[0], Not):
            return self.children[0].children[0]._to_normal_form()

        if isinstance(self.children[0], dc.And):
            return dc.Or(
                Not(self.children[0].children[0]), Not(self.children[0].children[1])
            )._to_normal_form()

        if isinstance(self.children[0], dc.Or):
            return dc.And(
                Not(self.children[0].children[0]), Not(self.children[0].children[1])
            )._to_normal_form()

        if isinstance(self.children[0], dc.Exists):
            return dc.Forall(
                self.children[0].variable, Not(self.children[0].children[0])
            )._to_normal_form()

        if isinstance(self.children[0], dc.Forall):
            return dc.Exists(
                self.children[0].variable, Not(self.children[0].children[0])
            )._to_normal_form()

        # if nothing changed, then convert child
        return Not(self.children[0]._to_normal_form())

    @typechecked
    def contains_variable(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable_typing(variable)

    @typechecked
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        return self.children[0].contains_variable_quantification(variable)

    @typechecked
    def _rename_variable(
        self, variable: dc.Variable, new_variable: dc.Variable
    ) -> dc.Formula:
        return Not(self.children[0]._rename_variable(variable, new_variable))

    @typechecked
    def _prune_tuple_atoms(self) -> dc.Formula | None:
        new_child = self.children[0]._prune_tuple_atoms()
        if new_child is None:
            return None

        return Not(new_child)

    @typechecked
    def get_used_variables(self) -> set[dc.Variable]:
        return self.children[0].get_used_variables()

    @typechecked
    def get_used_tuples(self) -> set[dc.Tuple]:
        return self.children[0].get_used_tuples()

    @typechecked
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        if self.children[0].contains_variable_typing(variable):
            # reminder: we are in Normal Form
            raise Exception("The declaration of a variable cannot be negated!")

        return self.children[0]._check_variable_legality(variable)



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class Or(tc.Formula):
    """
    A class representing the logical connective 'or'.
    """

    @typechecked
    def __init__(self, left: tc.Formula, right: tc.Formula) -> None:
        """
        Parameters
        ----------
        left : Formula
            Left subformula.
        right : Formula
            Right subformula.
        """
        super().__init__([left, right])

    @typechecked
    def __repr__(self) -> str:
        left = self.children[0]
        right = self.children[1]

        if isinstance(self.children[0], Or):
            left = (
                f"{self.children[0].children[0]} \\land {self.children[0].children[1]}"
            )
        elif not isinstance(self.children[0], tc.Not | tc.ATOM_TYPES | tc.Variable):
            left = f"({left})"

        if isinstance(self.children[1], Or):
            right = (
                f"{self.children[1].children[0]} \\land {self.children[1].children[1]}"
            )
        elif not isinstance(self.children[1], tc.Not | tc.ATOM_TYPES | tc.Variable):
            right = f"({right})"

        return f"{left} \\lor {right}"

    @typechecked
    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Or)
            and self.children[0] == other.children[0]
            and self.children[1] == other.children[1]
        )

    @typechecked
    def expand_quantifiers(self) -> tc.Formula:
        return Or(
            self.children[0].expand_quantifiers(), self.children[1].expand_quantifiers()
        )

    @typechecked
    def move_quantifiers_inwards(self) -> tc.Formula:
        return Or(
            self.children[0].move_quantifiers_inwards(),
            self.children[1].move_quantifiers_inwards(),
        )

    @typechecked
    def _move_same_type_quantifiers_outwards(
        self, move_exists: bool, move_forall: bool
    ) -> tc.Formula:
        if move_exists:
            if isinstance(self.children[0], tc.Exists):
                variable = self.children[0].variable
                if self.children[1].contains_variable(variable):
                    return Or(
                        self.children[0].rename_variable(variable), self.children[1]
                    )._move_same_type_quantifiers_outwards(
                        move_exists=True, move_forall=False
                    )
                else:
                    return tc.Exists(
                        variable,
                        Or(
                            self.children[0].children[0], self.children[1]
                        )._move_same_type_quantifiers_outwards(
                            move_exists=True, move_forall=False
                        ),
                    )
            if isinstance(self.children[1], tc.Exists):
                variable = self.children[1].variable
                if self.children[0].contains_variable(variable):
                    return Or(
                        self.children[0], self.children[1].rename_variable(variable)
                    )._move_same_type_quantifiers_outwards(
                        move_exists=True, move_forall=False
                    )
                else:
                    return tc.Exists(
                        variable,
                        Or(
                            self.children[0], self.children[1].children[0]
                        )._move_same_type_quantifiers_outwards(
                            move_exists=True, move_forall=False
                        ),
                    )

        if move_forall:
            if isinstance(self.children[0], tc.Forall):
                variable = self.children[0].variable
                if self.children[1].contains_variable(variable):
                    return Or(
                        self.children[0].rename_variable(variable), self.children[1]
                    )._move_same_type_quantifiers_outwards(
                        move_exists=False, move_forall=True
                    )
                else:
                    return tc.Forall(
                        variable,
                        Or(
                            self.children[0].children[0], self.children[1]
                        )._move_same_type_quantifiers_outwards(
                            move_exists=False, move_forall=True
                        ),
                    )
            if isinstance(self.children[1], tc.Forall):
                variable = self.children[1].variable
                if self.children[0].contains_variable(variable):
                    return Or(
                        self.children[0], self.children[1].rename_variable(variable)
                    )._move_same_type_quantifiers_outwards(
                        move_exists=False, move_forall=True
                    )
                else:
                    return tc.Forall(
                        variable,
                        Or(
                            self.children[0], self.children[1].children[0]
                        )._move_same_type_quantifiers_outwards(
                            move_exists=False, move_forall=True
                        ),
                    )

        return Or(
            self.children[0]._move_same_type_quantifiers_outwards(
                move_exists=move_exists, move_forall=move_forall
            ),
            self.children[1]._move_same_type_quantifiers_outwards(
                move_exists=move_exists, move_forall=move_forall
            ),
        )

    @typechecked
    def _to_sql(self) -> str:
        left = self.children[0]._to_sql()
        right = self.children[1]._to_sql()

        if left == "" and right == "":
            return ""
        if left == "" and right != "":
            return right
        if left != "" and right == "":
            return left

        if isinstance(self.children[0], Or):
            left = f"{self.children[0].children[0]._to_sql()} OR {self.children[0].children[1]._to_sql()}"
        elif not isinstance(self.children[0], tc.Not | tc.ATOM_TYPES):
            left = f"({left})"

        if isinstance(self.children[1], Or):
            right = f"{self.children[1].children[0]._to_sql()} OR {self.children[1].children[1]._to_sql()}"
        elif not isinstance(self.children[1], tc.Not | tc.ATOM_TYPES):
            right = f"({right})"

        return f"{left} OR {right}"

    @typechecked
    def flatten_quantifiers(self) -> tc.Formula:
        return Or(
            self.children[0].flatten_quantifiers(),
            self.children[1].flatten_quantifiers(),
        )

    @typechecked
    def _to_normal_form(self) -> tc.Formula:
        if isinstance(self.children[0], tc.Exists):
            self.children[0] = self.children[0]._to_normal_form()
            variable = self.children[0].variable
            if self.children[1].contains_variable(variable):
                return Or(
                    self.children[0].rename_variable(variable), self.children[1]
                )._to_normal_form()
            else:
                return tc.Exists(
                    variable,
                    Or(
                        self.children[0].children[0], self.children[1]
                    )._to_normal_form(),
                )

        if isinstance(self.children[0], tc.Forall):
            self.children[0] = self.children[0]._to_normal_form()
            variable = self.children[0].variable
            if self.children[1].contains_variable(variable):
                return Or(
                    self.children[0].rename_variable(variable), self.children[1]
                )._to_normal_form()
            else:
                return tc.Forall(
                    variable,
                    Or(
                        self.children[0].children[0], self.children[1]
                    )._to_normal_form(),
                )

        if isinstance(self.children[1], tc.Exists):
            self.children[1] = self.children[1]._to_normal_form()
            variable = self.children[1].variable
            if self.children[0].contains_variable(variable):
                return Or(
                    self.children[0], self.children[1].rename_variable(variable)
                )._to_normal_form()
            else:
                return tc.Exists(
                    variable,
                    Or(
                        self.children[0], self.children[1].children[0]
                    )._to_normal_form(),
                )

        if isinstance(self.children[1], tc.Forall):
            self.children[1] = self.children[1]._to_normal_form()
            variable = self.children[1].variable
            if self.children[0].contains_variable(variable):
                return Or(
                    self.children[0], self.children[1].rename_variable(variable)
                )._to_normal_form()
            else:
                return tc.Forall(
                    variable,
                    Or(
                        self.children[0], self.children[1].children[0]
                    )._to_normal_form(),
                )

        return Or(
            self.children[0]._to_normal_form(), self.children[1]._to_normal_form()
        )

    @typechecked
    def contains_variable(self, variable: tc.Variable) -> bool:
        return self.children[0].contains_variable(variable) or self.children[
            1
        ].contains_variable(variable)

    @typechecked
    def contains_variable_typing(self, variable: tc.Variable) -> bool:
        return self.children[0].contains_variable_typing(variable) or self.children[
            1
        ].contains_variable_typing(variable)

    @typechecked
    def _rename_variable(
        self, old_variable: tc.Variable, new_variable: tc.Variable
    ) -> tc.Formula:
        if self.children[0] == old_variable and self.children[1] == old_variable:
            return Or(new_variable, new_variable)

        if self.children[0] != old_variable and self.children[1] == old_variable:
            return Or(
                self.children[0]._rename_variable(old_variable, new_variable),
                new_variable,
            )

        if self.children[0] == old_variable and self.children[1] != old_variable:
            return Or(
                new_variable,
                self.children[1]._rename_variable(old_variable, new_variable),
            )

        return Or(
            self.children[0]._rename_variable(old_variable, new_variable),
            self.children[1]._rename_variable(old_variable, new_variable),
        )

    @typechecked
    def _prune_variable_atoms(self) -> tc.Formula | None:
        new_left = self.children[0]._prune_variable_atoms()
        new_right = self.children[1]._prune_variable_atoms()

        if new_left is None:
            return new_right

        if new_right is None:
            return new_left

        return Or(new_left, new_right)

    @typechecked
    def get_used_variables(self) -> set[tc.Variable]:
        return (
            self.children[0]
            .get_used_variables()
            .union(self.children[1].get_used_variables())
        )

    @typechecked
    def _check_variable_legality(self, variable: tc.Variable) -> bool:
        if self.children[0].contains_variable_typing(variable) and self.children[
            1
        ].contains_variable_typing(variable):
            # cannot happen due to renaming of variables
            raise Exception("Internal Error")

        if self.children[0].contains_variable_typing(variable) and self.children[
            1
        ].contains_variable(variable):
            raise Exception("Each variable must be declared before it is used")

        if self.children[0].contains_variable(variable) and self.children[
            1
        ].contains_variable_typing(variable):
            raise Exception("Each variable must be declared before it is used")

        return self.children[0]._check_variable_legality(variable) and self.children[
            1
        ]._check_variable_legality(variable)



from __future__ import annotations

from typeguard import typechecked

from typing import Any, Optional

import relational_calculus.domain_calculus as dc


class Tuple(dc.Formula):
    new_var_index = 1  # for the conversion of Non-Variables in Tuples to Variables

    @typechecked
    def __init__(
        self, type: str, variables: list[dc.Variable | dc.PRIMITIVE_TYPES | None]
    ) -> None:
        super().__init__([])
        if len([value for value in variables if value is not None]) == 0:
            raise Exception("Must have at least one variable")
        self.type = type.upper()
        self.variables = variables

    @typechecked
    def __repr__(self) -> str:
        output = f"\\text{{{self.type}}}("
        for variable in self.variables:
            if isinstance(variable, dc.Variable):
                output += f"\\text{{{variable.name}}}, "
            elif isinstance(variable, dc.PRIMITIVE_TYPES):
                output += f'\\text{{"{variable}"}}, '
            elif variable is None:
                output += f"\\_\\_, "

        output = output[: -len(", ")]  # remove last ', '
        output += ")"
        return output

    @typechecked
    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Tuple)
            and self.type == other.type
            and self.variables == other.variables
        )

    @typechecked
    def __hash__(self) -> int:
        return hash(
            tuple(
                (
                    Tuple,
                    self.type,
                    ", ".join(list(map(lambda var: str(var.name), self.variables))),
                )
            )
        )

    @typechecked
    def _to_sql(self) -> Optional[str]:
        # will be handled in DomainCalculus.to_sql()
        return ""

    @typechecked
    def _to_normal_form(self) -> dc.Formula:
        new_variables = dict()
        new_tuple = list()
        new_vars = list()
        for var in self.variables:
            if isinstance(var, dc.Variable):
                new_tuple.append(var)
            elif isinstance(var, dc.PRIMITIVE_TYPES):
                new_var = dc.Variable(f"{self.type}_{var}_{Tuple.new_var_index}")
                new_variables[new_var] = var
                new_tuple.append(new_var)
                new_vars.append(new_var)
                Tuple.new_var_index += 1
            elif var is None:
                new_var = dc.Variable(f"{self.type}_placeholder_{Tuple.new_var_index}")
                new_tuple.append(new_var)
                new_vars.append(new_var)
                Tuple.new_var_index += 1
        formula = Tuple(self.type, new_tuple)
        for new_var in new_vars:
            if new_var in new_variables.keys():
                formula = dc.Exists(
                    new_var, dc.And(formula, dc.Equals(new_var, new_variables[new_var]))
                )
            else:
                formula = dc.Exists(new_var, formula)
        return formula

    @typechecked
    def contains_variable(self, variable: dc.Variable) -> bool:
        return variable in self.variables

    @typechecked
    def contains_variable_typing(self, variable: dc.Variable) -> bool:
        return self.contains_variable(variable)

    @typechecked
    def contains_variable_quantification(self, variable: dc.Variable) -> bool:
        return False

    @typechecked
    def _rename_variable(
        self, old_variable: dc.Variable, new_variable: dc.Variable
    ) -> dc.Formula:
        new_tuple = list()
        for var in self.variables:
            if isinstance(var, dc.Variable):
                if var == old_variable:
                    new_tuple.append(new_variable)
                else:
                    new_tuple.append(var)
            elif isinstance(var, dc.PRIMITIVE_TYPES):
                new_tuple.append(var)
            elif var is None:
                new_tuple.append(None)
        return Tuple(self.type, new_tuple)

    @typechecked
    def _prune_tuple_atoms(self) -> dc.Formula | None:
        return None

    @typechecked
    def get_used_variables(self) -> set[dc.Variable]:
        return set(var for var in self.variables if isinstance(var, dc.Variable))

    @typechecked
    def get_used_tuples(self) -> set[dc.Tuple]:
        return {self}

    @typechecked
    def _check_variable_legality(self, variable: dc.Variable) -> bool:
        return True



from __future__ import annotations

from typeguard import typechecked

import relational_calculus.tuple_calculus as tc


class Variable(tc.Formula):
    """
    This class represents a variable of in tuple calculus, while technically also serving as an atom.
    """

    @typechecked
    def __init__(self, name: str, type: str) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the variable.
        type : str
            The relation / type that this variable is of.
        """
        super().__init__([])
        self.name = name.lower()
        self.type = type.upper()

    @typechecked
    def __repr__(self) -> str:
        return f"\\text{{{self.name}}} \\in \\text{{{self.type}}}"

    @typechecked
    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Variable)
            and self.name == other.name
            and self.type == other.type
        )

    @typechecked
    def __hash__(self) -> int:
        return hash((self.name, self.type))

    @typechecked
    def _to_sql(self) -> str:
        raise Exception("Internal Error: This should not happen!")

    @typechecked
    def _to_normal_form(self) -> tc.Formula:
        return self

    @typechecked
    def contains_variable(self, variable: tc.Variable) -> bool:
        return self == variable

    @typechecked
    def contains_variable_typing(self, variable: tc.Variable) -> bool:
        return self == variable

    @typechecked
    def _rename_variable(
        self, old_variable: tc.Variable, new_variable: tc.Variable
    ) -> tc.Formula:
        if self == old_variable:
            raise Exception("Internal Error")
        return self

    @typechecked
    def _prune_variable_atoms(self) -> tc.Formula | None:
        return None

    @typechecked
    def get_used_variables(self) -> set[tc.Variable]:
        return {self}

    @typechecked
    def _check_variable_legality(self, variable: tc.Variable) -> bool:
        return True

    @typechecked
    def __getitem__(self, attribute: str) -> tuple[Variable, str]:
        return self, attribute
